/*
Algorithm Name: Keccak
Authors: Guido Bertoni, Joan Daemen, Michaël Peeters and Gilles Van Assche
Date: October 27, 2008

This code, originally by Guido Bertoni, Joan Daemen, Michaël Peeters and
Gilles Van Assche as a part of the SHA-3 submission, is hereby put in the
public domain. It is given as is, without any guarantee.

For more information, feedback or questions, please refer to our website:
http://keccak.noekeon.org/
*/

#include "KeccakNISTInterface.h"
#include "KeccakPermutationInterface.h"

typedef unsigned char UINT8;
typedef unsigned long long int UINT64;

#define nrRounds 18
const UINT64 KeccakRoundConstants[nrRounds] = {
    0x0000000000000001ULL,
    0x0000000000008082ULL,
    0x800000000000808AULL,
    0x8000000080008000ULL,
    0x000000000000808BULL,
    0x0000000080000001ULL,
    0x8000000080008081ULL,
    0x8000000000008009ULL,
    0x000000000000008AULL,
    0x0000000000000088ULL,
    0x0000000080008009ULL,
    0x000000008000000AULL,
    0x000000008000808BULL,
    0x800000000000008BULL,
    0x8000000000008089ULL,
    0x8000000000008003ULL,
    0x8000000000008002ULL,
    0x8000000000000080ULL };

void KeccakPermutationOnWords(UINT64 *state);
void KeccakPermutationOnWordsAfterXoring8words(UINT64 *state, UINT64 *input);
void KeccakPermutationOnWordsAfterXoring16words(UINT64 *state, UINT64 *input);

void KeccakPermutation(unsigned char *state)
{
    // Little endianness assumed
    KeccakPermutationOnWords((UINT64*)state);
}

void KeccakPermutationAfterXor(unsigned char *state, const unsigned char *data, unsigned int dataLengthInBytes)
{
    unsigned int i, j;

    // Little endianness assumed
    if (dataLengthInBytes == 128)
        KeccakPermutationOnWordsAfterXoring16words((UINT64*)state, (UINT64*)data);
    else if (dataLengthInBytes == 64)
        KeccakPermutationOnWordsAfterXoring8words((UINT64*)state, (UINT64*)data);
    else {
        for(j=0; j<(dataLengthInBytes/8); j++)
            ((UINT64*)state)[j] ^= ((UINT64*)data)[j];
        for(i=j*8; i<dataLengthInBytes; i++)
            state[i] ^= data[i];
        KeccakPermutationOnWords((UINT64*)state);
    }
}

void KeccakInitialize()
{
}

#define declareABCD \
    UINT64 Aba, Abe, Abi, Abo, Abu; \
    UINT64 Aga, Age, Agi, Ago, Agu; \
    UINT64 Aka, Ake, Aki, Ako, Aku; \
    UINT64 Ama, Ame, Ami, Amo, Amu; \
    UINT64 Asa, Ase, Asi, Aso, Asu; \
    UINT64 Bba, Bbe, Bbi, Bbo, Bbu; \
    UINT64 Bga, Bge, Bgi, Bgo, Bgu; \
    UINT64 Bka, Bke, Bki, Bko, Bku; \
    UINT64 Bma, Bme, Bmi, Bmo, Bmu; \
    UINT64 Bsa, Bse, Bsi, Bso, Bsu; \
    UINT64 Ca, Ce, Ci, Co, Cu; \
    UINT64 Da, De, Di, Do, Du;

#define ROL64(a, offset) ((((UINT64)a) << offset) ^ (((UINT64)a) >> (64-offset)))

#define round(i) \
    Ca = Aba^Aga^Aka^Ama^Asa; \
    Ce = Abe^Age^Ake^Ame^Ase; \
    Cu = Abu^Agu^Aku^Amu^Asu; \
    Da = Cu^ROL64(Ce, 1); \
    Aba ^= Da; \
    Bba = Aba; \
    Aga ^= Da; \
    Bme = ROL64(Aga, 36); \
    Aka ^= Da; \
    Bgi = ROL64(Aka,  3); \
    Ama ^= Da; \
    Bso = ROL64(Ama, 41); \
    Asa ^= Da; \
    Bku = ROL64(Asa, 18); \
    Ci = Abi^Agi^Aki^Ami^Asi; \
    Co = Abo^Ago^Ako^Amo^Aso; \
    De = Ca^ROL64(Ci, 1); \
    Abe ^= De; \
    Bka = ROL64(Abe,  1); \
    Age ^= De; \
    Bbe = ROL64(Age, 44); \
    Ake ^= De; \
    Bmi = ROL64(Ake, 10); \
    Ame ^= De; \
    Bgo = ROL64(Ame, 45); \
    Ase ^= De; \
    Bsu = ROL64(Ase,  2); \
    Di = Ce^ROL64(Co, 1); \
    Abi ^= Di; \
    Bsa = ROL64(Abi, 62); \
    Agi ^= Di; \
    Bke = ROL64(Agi,  6); \
    Aki ^= Di; \
    Bbi = ROL64(Aki, 43); \
    Ami ^= Di; \
    Bmo = ROL64(Ami, 15); \
    Asi ^= Di; \
    Bgu = ROL64(Asi, 61); \
    Do = Ci^ROL64(Cu, 1); \
    Abo ^= Do; \
    Bga = ROL64(Abo, 28); \
    Ago ^= Do; \
    Bse = ROL64(Ago, 55); \
    Ako ^= Do; \
    Bki = ROL64(Ako, 25); \
    Amo ^= Do; \
    Bbo = ROL64(Amo, 21); \
    Aso ^= Do; \
    Bmu = ROL64(Aso, 56); \
    Du = Co^ROL64(Ca, 1); \
    Abu ^= Du; \
    Bma = ROL64(Abu, 27); \
    Agu ^= Du; \
    Bge = ROL64(Agu, 20); \
    Aku ^= Du; \
    Bsi = ROL64(Aku, 39); \
    Asa = Bsa ^ ((~Bse)&Bsi); \
    Ase = Bse ^ ((~Bsi)&Bso); \
    Amu ^= Du; \
    Bko = ROL64(Amu,  8); \
    Asu ^= Du; \
    Bbu = ROL64(Asu, 14); \
    Asi = Bsi ^ ((~Bso)&Bsu); \
    Aso = Bso ^ ((~Bsu)&Bsa); \
    Asu = Bsu ^ ((~Bsa)&Bse); \
    Ake = Bke ^ ((~Bki)&Bko); \
    Ame = Bme ^ ((~Bmi)&Bmo); \
    Aka = Bka ^ ((~Bke)&Bki); \
    Ama = Bma ^ ((~Bme)&Bmi); \
    Aki = Bki ^ ((~Bko)&Bku); \
    Ami = Bmi ^ ((~Bmo)&Bmu); \
    Ako = Bko ^ ((~Bku)&Bka); \
    Amo = Bmo ^ ((~Bmu)&Bma); \
    Aku = Bku ^ ((~Bka)&Bke); \
    Amu = Bmu ^ ((~Bma)&Bme); \
    Abe = Bbe ^ ((~Bbi)&Bbo); \
    Age = Bge ^ ((~Bgi)&Bgo); \
    Abi = Bbi ^ ((~Bbo)&Bbu); \
    Agi = Bgi ^ ((~Bgo)&Bgu); \
    Abo = Bbo ^ ((~Bbu)&Bba); \
    Ago = Bgo ^ ((~Bgu)&Bga); \
    Abu = Bbu ^ ((~Bba)&Bbe); \
    Agu = Bgu ^ ((~Bga)&Bge); \
    Aba = Bba ^ ((~Bbe)&Bbi); \
    Aga = Bga ^ ((~Bge)&Bgi); \
    Aba ^= KeccakRoundConstants[i]; \

#define copyFromState(X, state) \
    X##ba = state[ 0]; \
    X##be = state[ 1]; \
    X##bi = state[ 2]; \
    X##bo = state[ 3]; \
    X##bu = state[ 4]; \
    X##ga = state[ 5]; \
    X##ge = state[ 6]; \
    X##gi = state[ 7]; \
    X##go = state[ 8]; \
    X##gu = state[ 9]; \
    X##ka = state[10]; \
    X##ke = state[11]; \
    X##ki = state[12]; \
    X##ko = state[13]; \
    X##ku = state[14]; \
    X##ma = state[15]; \
    X##me = state[16]; \
    X##mi = state[17]; \
    X##mo = state[18]; \
    X##mu = state[19]; \
    X##sa = state[20]; \
    X##se = state[21]; \
    X##si = state[22]; \
    X##so = state[23]; \
    X##su = state[24];

#define copyToState(state, X) \
    state[ 0] = X##ba; \
    state[ 1] = X##be; \
    state[ 2] = X##bi; \
    state[ 3] = X##bo; \
    state[ 4] = X##bu; \
    state[ 5] = X##ga; \
    state[ 6] = X##ge; \
    state[ 7] = X##gi; \
    state[ 8] = X##go; \
    state[ 9] = X##gu; \
    state[10] = X##ka; \
    state[11] = X##ke; \
    state[12] = X##ki; \
    state[13] = X##ko; \
    state[14] = X##ku; \
    state[15] = X##ma; \
    state[16] = X##me; \
    state[17] = X##mi; \
    state[18] = X##mo; \
    state[19] = X##mu; \
    state[20] = X##sa; \
    state[21] = X##se; \
    state[22] = X##si; \
    state[23] = X##so; \
    state[24] = X##su;

void KeccakPermutationOnWords(UINT64 *state)
{
    declareABCD
    unsigned int i;

    copyFromState(A, state);
    for(i=0; i<nrRounds; i+=3) {
        round(i);
        round(i+1);
        round(i+2);
    }
    copyToState(state, A);
}

#define copyFromStateAndXor8words(X, state, input) \
    X##ba = state[ 0]^input[ 0]; \
    X##be = state[ 1]^input[ 1]; \
    X##bi = state[ 2]^input[ 2]; \
    X##bo = state[ 3]^input[ 3]; \
    X##bu = state[ 4]^input[ 4]; \
    X##ga = state[ 5]^input[ 5]; \
    X##ge = state[ 6]^input[ 6]; \
    X##gi = state[ 7]^input[ 7]; \
    X##go = state[ 8]; \
    X##gu = state[ 9]; \
    X##ka = state[10]; \
    X##ke = state[11]; \
    X##ki = state[12]; \
    X##ko = state[13]; \
    X##ku = state[14]; \
    X##ma = state[15]; \
    X##me = state[16]; \
    X##mi = state[17]; \
    X##mo = state[18]; \
    X##mu = state[19]; \
    X##sa = state[20]; \
    X##se = state[21]; \
    X##si = state[22]; \
    X##so = state[23]; \
    X##su = state[24];

void KeccakPermutationOnWordsAfterXoring8words(UINT64 *state, UINT64 *input)
{
    declareABCD
    unsigned int i;

    copyFromStateAndXor8words(A, state, input);
    for(i=0; i<nrRounds; i+=3) {
        round(i);
        round(i+1);
        round(i+2);
    }
    copyToState(state, A);
}

#define copyFromStateAndXor16words(X, state, input) \
    X##ba = state[ 0]^input[ 0]; \
    X##be = state[ 1]^input[ 1]; \
    X##bi = state[ 2]^input[ 2]; \
    X##bo = state[ 3]^input[ 3]; \
    X##bu = state[ 4]^input[ 4]; \
    X##ga = state[ 5]^input[ 5]; \
    X##ge = state[ 6]^input[ 6]; \
    X##gi = state[ 7]^input[ 7]; \
    X##go = state[ 8]^input[ 8]; \
    X##gu = state[ 9]^input[ 9]; \
    X##ka = state[10]^input[10]; \
    X##ke = state[11]^input[11]; \
    X##ki = state[12]^input[12]; \
    X##ko = state[13]^input[13]; \
    X##ku = state[14]^input[14]; \
    X##ma = state[15]^input[15]; \
    X##me = state[16]; \
    X##mi = state[17]; \
    X##mo = state[18]; \
    X##mu = state[19]; \
    X##sa = state[20]; \
    X##se = state[21]; \
    X##si = state[22]; \
    X##so = state[23]; \
    X##su = state[24];

void KeccakPermutationOnWordsAfterXoring16words(UINT64 *state, UINT64 *input)
{
    declareABCD
    unsigned int i;

    copyFromStateAndXor16words(A, state, input);
    for(i=0; i<nrRounds; i+=3) {
        round(i);
        round(i+1);
        round(i+2);
    }
    copyToState(state, A);
}
